/* 
// AUTHOR
// N. Nielsen
//
// LICENSE
// This software is in the public domain.
//
// The software is provided "as is", without warranty of any kind,
// express or implied, including but not limited to the warranties
// of merchantability, fitness for a particular purpose, and
// noninfringement. In no event shall the author(s) be liable for any
// claim, damages, or other liability, whether in an action of
// contract, tort, or otherwise, arising from, out of, or in connection
// with the software or the use or other dealings in the software.
// 
// SUPPORT
// Send bug reports to: <nielsen@memberwebs.com>
//
// CHANGES
// 0.5.1	kvm_openfiles uses /dev/null
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/sysctl.h>
#include <sys/proc.h>
#include <sys/user.h>

#include <paths.h>  
#include <signal.h>
#include <stdio.h>
#include <syslog.h>
#include <err.h>
#include <errno.h>
#include <stdarg.h>
#include <unistd.h>
#include <kvm.h>
#include <limits.h>
#include <fcntl.h>

#include "getjail.h"

#ifdef HAVE_CONFIG_H 
#include "../config.h"
#endif

/* The big long stop process */
static int stopJail(char* jailName);

/* Signals the jailer for various requests */
static int signalJail(char* jailName, int signal);

static void killProcesses(pid_t* pids, int signal);
static int getJailProcesses(const char* jailName, pid_t* pidJailer, pid_t** pids);

static void usage();


/* The timeout to wait between kills */
#define		DEFAULT_TIMEOUT		10
int g_timeout = DEFAULT_TIMEOUT;

/* To find the jailer process look for this command */
#define		JAILER_COMMAND		"jailer"


int main(int argc, char* argv[])
{
	/* If this gets set then only signal jailer, no kill */
	int signal = 0;
	int ch = 0;

	while((ch = getopt(argc, argv, "hrt:")) != -1)
	{
		switch(ch)
		{
		/* Send halt request to jailer */
		case 'h':
			signal = SIGQUIT;
			break;

		/* Send restart request to jailer */
		case 'r':
			signal = SIGHUP;
			break;

		/* Timeout to use between kills */
		case 't':
			g_timeout = atoi(optarg);
			if(g_timeout <= 0)
				errx(1, "invalid timeout argument: %s", optarg);
			break;

		case '?':
		default:
			usage();
		}
	}

	argc -= optind;
	argv += optind;

	/* Make sure we have a jailName */
	if(argc == 0)
		usage();

	/* For each jail */
	while(argc > 0)
	{
		/* If a signal option was set above then signal,
		   otherwise kill */
		if(signal == 0)
			stopJail(*argv);
		else
			signalJail(*argv, signal);

		argc--;
		argv++;
	}

	return 0;
}


int signalJail(char* jailName, int signal)
{
	pid_t jailerPid = 0;

	/* Only ask for jailer pid */
	if(getJailProcesses(jailName, &jailerPid, NULL) < 0)
		return -1;

	if(jailerPid == 0)
		errx(1, "%s: jailer not running in jail", jailName);

	if(kill(jailerPid, signal) < 0)
		err(1, "%s: couldn't signal jailer");
}


int stopJail(char* jailName)
{
	pid_t jailerPid = 0;
	pid_t* jailProcesses = NULL;
	int pass = 0;

	/* 
	 * Multiple passes are used to do different things.
	 * Each time the jails processes are listed.
	 */
	while(getJailProcesses(jailName, &jailerPid, &jailProcesses))
	{
		/* First pass is killing the jailer */
		if(pass == 0)
		{
			if(jailerPid == 0)
			{
				/* No jailer */
				warnx("%s: jailer not running in jail", jailName);
			}

			else
			{
				if(kill(jailerPid, SIGTERM) < 0)
				{
					if(errno == ESRCH)
						break;

					err(1, "%s: couldn't signal jailer:", jailName);
				}
			}
		}

		/* Okay now we wait for the jailer to exit */
		else if(pass >= 1 && pass <= g_timeout)
		{
			/* If jailer is gone then skip to pass 5 */
			if(jailerPid == 0)
				pass = g_timeout;

			/* If we get here, jailer looks like it's irresponsive */
			else if(pass == g_timeout && jailerPid != 0)
				warnx("%s: jailer (pid %d) won't quit. terminating jail...", jailName, jailerPid);

			sleep(1);
		}

		/* next pass is sending all remaining processes a TERM */
		else if(pass == (g_timeout + 1))
		{
			killProcesses(jailProcesses, SIGTERM);
		}

		/* 7rd to 13th pass is waiting for the above to exit */
		else if(pass >= (g_timeout + 2) && pass <= ((g_timeout * 2) + 2))
		{
			sleep(1);
		}

		/* And if that didn't do it, well then give up */
		else if(pass >= ((g_timeout * 2) + 3))
		{
			errx(1, "%s: couldn't stop jail, processes wouldn't die", jailName);
			break;
		}

		if(jailProcesses)
			free(jailProcesses);

		pass++;
	}

	return 0;
}

void killProcesses(pid_t* pids, int signal)
{
	/* Note that we assume pids is null terminated
	   this is what getJailProcesses returns */

	while(*pids)
	{
		if(kill(*pids, signal) < 0)
		{
			/* We ignore missing process errors */
			if(errno != ESRCH)
				err(1, "couldn't kill process: %d", *pids);
		}

		pids++;
	}
}

int getJailProcesses(const char* jailName, pid_t* pidJailer, pid_t** pids)
{
	kvm_t *kd;
	struct kinfo_proc* kp;
	char errbuf[_POSIX2_LINE_MAX];
	char pidJail[JAIL_BUFF_SIZE];
	int nentries, i, j;

	/* Open the kernel interface */
	kd = kvm_openfiles(_PATH_DEVNULL, _PATH_DEVNULL, _PATH_DEVNULL, 
					   O_RDONLY, errbuf);
	if(kd == 0)
		errx(1, "%s", errbuf);

	/* Get a process listing */
	if((kp = kvm_getprocs(kd, KERN_PROC_ALL, 0, &nentries)) == 0)
		errx(1, "%s", kvm_geterr(kd));

	/* Allocate memory */
	if(pids)
	{
		if((*pids = (pid_t*)malloc((nentries + 1) * sizeof(pid_t))) == NULL)
			err(1, "out of memory");
	}

	/* Okay now loop and look at each process' jail */
	for(i = 0, j = 0; i < nentries; i++)
	{
		struct proc* p = &(kp[i].kp_proc);

		/* Check the flags first */
		if(!p || !p->p_prison || !(p->p_flag & P_JAILED))
			continue;

		/* Now actually get the jail name */
		if(getpidjail(p->p_pid, pidJail) < 0)
			continue;

		if(strcmp(pidJail, jailName))
			continue;

		/* Copy the PID over */
		if(pids)
			(*pids)[j++] = p->p_pid;

		/* If it's the jailer then copy that */
		if(pidJailer)
		{
			if(strstr(p->p_comm, JAILER_COMMAND))
				*pidJailer = p->p_pid;
		}
		
	}

	/* Null terminate pids array */
	if(pids)
		(*pids)[j] = 0;

	kvm_close(kd);

	return j == 0 ? 0 : 1;
}



static void usage()
{
	fprintf(stderr, "usage: killjail [ -h | -r ] [ -t timeout ] jailname ...\n");
	exit(2);
}

